/*
 * Copyright (c) 2009, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * The names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#ifndef DICT_H
#define DICT_H

#include <stdint.h>
#include <vector>
#include <string>

#ifndef SPTH_CL
#  include <QObject>
#  include <QWidget>
#endif /* !SPTH_CL */

#include "defines.h"

using std::vector;
using std::string;

namespace Spellathon
{
	struct WordEnt {
		char	word[MAX_LETRS+1];
		uint8_t	len;
	};

	struct DictEnt {	
		char	sign[MAX_LETRS+1];
		WordEnt	*wordp;
	};

	struct PrecompEnt {
		char	sign[MAX_LETRS+1];
		uint16_t nsol;
	};
	
#ifdef SPTH_CL
	class Dict {
	public:
		Dict(); 
#else		
	class Dict : public QObject {
		Q_OBJECT
	public:
		Dict(QWidget *p = 0); 
#endif /* SPTH_CL */
		~Dict() { destroy(); }

		void		destroy();
		void		init(const char *path, bool precomp = PRECOMP, bool pluralSieve = PLURAL_SIEVE);
		void		matchSign(const char *sign, char *w[], int *idx, int *has7) const;
		int			off7() const { return dictLst_.lenOff[7-4].start; } 
		int			nent7() const { return dictLst_.lenOff[7-4].end - dictLst_.lenOff[7-4].start + 1; } 
		int			precompSize() const { return precompHdr_.precompLst_size; }
		bool		precomp() const { return precomp_; }
		string		dictPath() const { return dictPath_; }
		bool		pluralSieve() const { return pluralSieve_; }
		const PrecompEnt*	precompArr() const { return precompLst_.arr; }
		const DictEnt*		dictArr() const { return dictLst_.arr; }

	private:
		inline int	bsearch(const char * sign, int slen) const;
		void		dictOpen(const char *path);
		void		precompOpen(const char *path);
		void		precompCreate(const char *path);
		void		setPrecompPath();
		void		processNonPlural();
		void		dictProcess();
		inline bool	isNotPlural(const char *word, int slen) const;
		bool		matchWord(const char *word, int slen) const;

	private:
		enum {MAX_LNLEN = 64, MAX_PATH_LEN = 1024};
		enum {PRECOMP_VER = 4};

		struct Offset {
			int start;	
			int end; 
		};

		struct {
			DictEnt		*arr;	
			int			size;	
			Offset		lenOff[4];	
			Offset		idxOff[4][26][26][26];	
		} dictLst_;

		struct {
			WordEnt		*arr;	/* list of all words of 3 <= strlen() <=7	 */
								/* the 3 letter words are needed for finding */
								/* plurals.						 			 */
			int			size;
			Offset		lenOff[5];	
			Offset		idxOff[5][26][26];	
		} wordLst_;

		struct {
			PrecompEnt	*arr;	
			int			size;
		} precompLst_;

		struct {
			char	magic[12];
			int		version;
			int		wordLst_size;	
			int		precompLst_size;
			int		pluralSieve;
			Offset	wordLst_lenOff[5];	
			Offset	wordLst_idxOff[5][26][26];	
			char	dictPath[MAX_PATH_LEN];
		} precompHdr_;

		struct {
			WordEnt		**arr;
			int			size;
			Offset		lenOff[4];	
			Offset		idxOff[4][26][26];	
		} nonPluralLst_;

		bool	precomp_;
		bool	pluralSieve_;
		char	dictbinPath_[MAX_PATH_LEN];
		string	dictPath_;
#ifndef SPTH_CL
		QWidget	*gui_h_;
#endif /* SPTH_CL */

		Dict(const Dict &);
		Dict & operator=(const Dict &);
	};
}

#endif /* DICT_H */
